/*
 * Copyright (C) 2009 - 2012 Robin Seggelmann, seggelmann@fh-muenster.de,
 *                           Michael Tuexen, tuexen@fh-muenster.de
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "aqua.h"
#include "dtls.h"

#define BUFFER_SIZE          (1<<16)
#define COOKIE_SECRET_LENGTH 16

int verbose	= 1;
int veryverbose	= 1;
unsigned char cookie_secret[COOKIE_SECRET_LENGTH];
int cookie_initialized	= 0;

extern FILE *			log_file;
extern struct state *		state;

#if WIN32
static HANDLE* mutex_buf = NULL;
#else
static pthread_mutex_t* mutex_buf = NULL;
#endif

void
locking_function(int mode, int n, const char *file _unused, int line _unused)
{
	if (mode & CRYPTO_LOCK)
#ifdef WIN32
		WaitForSingleObject(mutex_buf[n], INFINITE);
	else
		ReleaseMutex(mutex_buf[n]);
#else
		pthread_mutex_lock(&mutex_buf[n]);
	else
		pthread_mutex_unlock(&mutex_buf[n]);
#endif
}

unsigned long
id_function(void) {
#ifdef WIN32
	return (unsigned long) GetCurrentThreadId();
#else
	return (unsigned long) pthread_self();
#endif
}

int
THREAD_setup()
{
	int i;

#ifdef WIN32
	mutex_buf = (HANDLE*) malloc(CRYPTO_num_locks() * sizeof(HANDLE));
#else
	mutex_buf = (pthread_mutex_t*) malloc(CRYPTO_num_locks() * sizeof(pthread_mutex_t));
#endif
	if (!mutex_buf)
		return 0;
	for (i = 0; i < CRYPTO_num_locks(); i++)
#ifdef WIN32
		mutex_buf[i] = CreateMutex(NULL, FALSE, NULL);
#else
		pthread_mutex_init(&mutex_buf[i], NULL);
#endif
	CRYPTO_set_id_callback(id_function);
	CRYPTO_set_locking_callback(locking_function);
	return 1;
}

int THREAD_cleanup()
{
	int i;

	if (!mutex_buf)
		return 0;

	CRYPTO_set_id_callback(NULL);
	CRYPTO_set_locking_callback(NULL);
	for (i = 0; i < CRYPTO_num_locks(); i++)
#ifdef WIN32
	CloseHandle(mutex_buf[i]);
#else
	pthread_mutex_destroy(&mutex_buf[i]);
#endif
	free(mutex_buf);
	mutex_buf = NULL;
	return 1;
}

int
handle_socket_error()
{
	switch (errno) {
		case EINTR:
			/* Interrupted system call.
			 * Just ignore.
			 */
			LOG("Interrupted system call!\n");
			return 1;
		case EBADF:
			/* Invalid socket.
			 * Must close connection.
			 */
			LOG("Invalid socket!\n");
			return 0;
			break;
#ifdef EHOSTDOWN
		case EHOSTDOWN:
			/* Host is down.
			 * Just ignore, might be an attacker
			 * sending fake ICMP messages.
			 */
			LOG("Host is down!\n");
			return 1;
#endif
#ifdef ECONNRESET
		case ECONNRESET:
			/* Connection reset by peer.
			 * Just ignore, might be an attacker
			 * sending fake ICMP messages.
			 */
			LOG("Connection reset by peer!\n");
			return 1;
#endif
		case ENOMEM:
			/* Out of memory.
			 * Must close connection.
			 */
			LOG("Out of memory!\n");
			return 0;
			break;
		case EACCES:
			/* Permission denied.
			 * Just ignore, we might be blocked
			 * by some firewall policy. Try again
			 * and hope for the best.
			 */
			LOG("Permission denied!\n");
			return 1;
			break;
		default:
			/* Something unexpected happened */
			LOG("Unexpected error! (errno = %d)\n", errno);
			return 0;
			break;
	}
	return 0;
}

int
generate_cookie(SSL *ssl, unsigned char *cookie, unsigned int *cookie_len)
{
	unsigned char *buffer, result[EVP_MAX_MD_SIZE];
	unsigned int length = 0, resultlength;
	addr peer;

	/* Initialize a random secret */
	if (!cookie_initialized)
		{
		if (!RAND_bytes(cookie_secret, COOKIE_SECRET_LENGTH))
			{
			LOG("error setting random cookie secret\n");
			return 0;
			}
		cookie_initialized = 1;
		}

	/* Read peer information */
	(void) BIO_dgram_get_peer(SSL_get_rbio(ssl), &peer);

	/* Create buffer with peer's address and port */
	length = 0;
	switch (peer.ss.ss_family) {
		case AF_INET:
			length += sizeof(struct in_addr);
			break;
		case AF_INET6:
			LOG("BIG NO");
			length += sizeof(struct in6_addr);
			break;
		default:
			OPENSSL_assert(0);
			break;
	}
	length += sizeof(in_port_t);
	buffer = (unsigned char*) OPENSSL_malloc(length);

	if (buffer == NULL)
		{
		LOG("out of memory\n");
		return 0;
		}

	switch (peer.ss.ss_family) {
		case AF_INET:
			memcpy(buffer,
			       &peer.s4.sin_port,
			       sizeof(in_port_t));
			memcpy(buffer + sizeof(peer.s4.sin_port),
			       &peer.s4.sin_addr,
			       sizeof(struct in_addr));
			break;
		case AF_INET6:
			LOG("BIG NO");
			memcpy(buffer,
			       &peer.s6.sin6_port,
			       sizeof(in_port_t));
			memcpy(buffer + sizeof(in_port_t),
			       &peer.s6.sin6_addr,
			       sizeof(struct in6_addr));
			break;
		default:
			OPENSSL_assert(0);
			break;
	}

	/* Calculate HMAC of buffer using the secret */
	HMAC(EVP_sha1(), (const void*) cookie_secret, COOKIE_SECRET_LENGTH,
	     (const unsigned char*) buffer, length, result, &resultlength);
	OPENSSL_free(buffer);

	memcpy(cookie, result, resultlength);
	*cookie_len = resultlength;

	return 1;
}

int
verify_cookie(SSL *ssl, unsigned char *cookie, unsigned int cookie_len)
{
	unsigned char *buffer, result[EVP_MAX_MD_SIZE];
	unsigned int length = 0, resultlength;
	addr peer;

	/* If secret isn't initialized yet, the cookie can't be valid */
	if (!cookie_initialized)
		return 0;

	/* Read peer information */
	(void) BIO_dgram_get_peer(SSL_get_rbio(ssl), &peer);

	/* Create buffer with peer's address and port */
	length = 0;
	switch (peer.ss.ss_family) {
		case AF_INET:
			length += sizeof(struct in_addr);
			break;
		case AF_INET6:
			LOG("BIG NO");
			length += sizeof(struct in6_addr);
			break;
		default:
			OPENSSL_assert(0);
			break;
	}
	length += sizeof(in_port_t);
	buffer = (unsigned char*) OPENSSL_malloc(length);

	if (buffer == NULL)
		{
		LOG("out of memory\n");
		return 0;
		}

	switch (peer.ss.ss_family) {
		case AF_INET:
			memcpy(buffer,
			       &peer.s4.sin_port,
			       sizeof(in_port_t));
			memcpy(buffer + sizeof(in_port_t),
			       &peer.s4.sin_addr,
			       sizeof(struct in_addr));
			break;
		case AF_INET6:
			LOG("BIG NO");
			memcpy(buffer,
			       &peer.s6.sin6_port,
			       sizeof(in_port_t));
			memcpy(buffer + sizeof(in_port_t),
			       &peer.s6.sin6_addr,
			       sizeof(struct in6_addr));
			break;
		default:
			OPENSSL_assert(0);
			break;
	}

	/* Calculate HMAC of buffer using the secret */
	HMAC(EVP_sha1(), (const void*) cookie_secret, COOKIE_SECRET_LENGTH,
	     (const unsigned char*) buffer, length, result, &resultlength);
	OPENSSL_free(buffer);

	if (cookie_len == resultlength && memcmp(result, cookie, resultlength) == 0)
		return 1;

	return 0;
}

struct pass_info {
	addr server_addr, client_addr;
	SSL *ssl;
#if WIN32
	DWORD tid;
#else
	pthread_t tid;
#endif
};

int
dtls_verify_callback(int ok _unused, X509_STORE_CTX *ctx _unused)
{
	/* This function should ask the user
	 * if he trusts the received certificate.
	 * Here we always trust.
	 */
	return 1;
}


#ifdef WIN32
DWORD WINAPI
rated_write(LPVOID *info)
#else
void *
rated_write(void *			arg)
#endif
{
	struct node_info *		node	= arg;
	SSL *				ssl	= node->ssl;
	char				null_buf[AQUA_PACKET_SIZE];
	int				len	= AQUA_PACKET_SIZE;
	int				padding_len	= AQUA_PACKET_SIZE;
	char *				to_send;

	if (state->role == AQUA_APP_PROXY && node->role == AQUA_SP)
		padding_len	= AQUA_SP_CLIENT_PACKET_SIZE;

	LOG("FIXME rated write created on %i", node->index);

	while (1)
	{
		// add timestamp here & end so we can remove that from sleep time.
		MUTEX_LOCK(&node->ssl_lock);

		if (SSL_get_shutdown(ssl) & SSL_RECEIVED_SHUTDOWN)
		{
			MUTEX_UNLOCK(&node->ssl_lock);
			break;
		}

		// tmp FIXME padding
		if (padding_len == AQUA_SP_CLIENT_PACKET_SIZE && !node->queue[node->queue_head]) // FIXME we should be creating padding on a separate thread, this is bad
		{
			//send_cell_to_sp(NULL, node->secret, 0, node->index); // FIXME tmp
			send_cell_to_sp(null_buf, node->secret, 1, node->index); // FIXME null = secret. maybe we should put it in the node?
		}
		// end tmp FIXME

		MUTEX_LOCK(&node->queue_lock);
		if (!(to_send = (char *) node->queue[node->queue_head]))
		{
			to_send	= null_buf;
			len	= AQUA_PACKET_SIZE; //FIXME: not in the case of SP
		}
		else
			len	= node->queue_len[node->queue_head];

		//if (to_send != null_buf) // removes chaffing, can be useful for debugging. death
		len	= SSL_write(ssl, to_send, len);

		if (to_send != null_buf)
		{
			//LOG("sent %i", len);
			//free(to_send);
			free(node->queue[node->queue_head]);
			//LOG("mem node %i, free %i", node->index, node->queue_head);
			node->queue[node->queue_head]	= NULL;
			if (++node->queue_head == MAX_QUEUE)
				node->queue_head	= 0;
			if (!node->queue[node->queue_head]) // head overtakes tail when queue is empty
			{
				node->queue_head	= 0;
				node->queue_tail	= 0;
			}
		}
		MUTEX_UNLOCK(&node->queue_lock);

		switch (SSL_get_error(ssl, len))
		{
			case SSL_ERROR_NONE:
				//if (verbose == 2) // FIXME killing this for now
				//	LOG("Thread %lx: wrote %d bytes\n", id_function(), (int) len);

				break;
			case SSL_ERROR_WANT_WRITE:
				/* Can't write because of a renegotiation, so
				 * we actually have to retry sending this message...
				 * FIXME: see SSL_MODE_AUTO_RETRY flag of the SSL_CTX_set_mode
				 */
				break;
			case SSL_ERROR_WANT_READ:
				/* continue with reading */
				break;
			case SSL_ERROR_SYSCALL:
				LOG("Socket write error: ");
				if (!handle_socket_error())
					goto cleanup;
				break;
			case SSL_ERROR_SSL:
				LOG("SSL write error: ");
				LOG("%s (%d)\n", ERR_error_string(ERR_get_error(), null_buf), SSL_get_error(ssl, len));
				goto cleanup;
				break;
			default:
				LOG("Unexpected error while writing!\n");
				goto cleanup;
				break;
		}
		MUTEX_UNLOCK(&node->ssl_lock);
		usleep(AQUA_CHAFFING_TIME); // FIXME put this in a per node setting.
	}

cleanup:
	// FIXME call free_node.
	MUTEX_UNLOCK(&node->ssl_lock);
	MUTEX_DESTROY(&node->ssl_lock);
	MUTEX_DESTROY(&node->queue_lock);
//	MUTEX_DESTROY(&node->aqua_lock);
	send_rm_node(node);
	node->used	= 0;
	for (int i = 0; i < MAX_QUEUE; ++i)
		if (node->queue[i])
		{
			free(node->queue[i]);
			node->queue[i] = NULL;
		}

	// FIXME, clean some aqua things too. completely kill conn?
	pthread_exit(NULL);
}


#ifdef WIN32
DWORD WINAPI
connection_handle(LPVOID *info) {
#else
void *
connection_handle(void *info)
{
#endif
	ssize_t len;
	char buf[AQUA_PACKET_SIZE * 2];
	char addrbuf[INET6_ADDRSTRLEN];
	struct pass_info *pinfo = (struct pass_info*) info;
	SSL *ssl = pinfo->ssl;
	int fd, reading = 0, ret;
	const int on = 1, off = 0;
	struct timeval timeout;
 	int num_timeouts = 0, max_timeouts = 5;

#ifndef WIN32
	pthread_detach(pthread_self());
#endif

	OPENSSL_assert(pinfo->client_addr.ss.ss_family == pinfo->server_addr.ss.ss_family);
	fd = socket(pinfo->client_addr.ss.ss_family, SOCK_DGRAM, 0);
	if (fd < 0) {
		perror("socket");
		goto cleanup;
	}

#ifdef WIN32
	setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (const char*) &on, (socklen_t) sizeof(on));
#else
	setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (const void*) &on, (socklen_t) sizeof(on));
#ifdef SO_REUSEPORT
	setsockopt(fd, SOL_SOCKET, SO_REUSEPORT, (const void*) &on, (socklen_t) sizeof(on));
#endif
#endif
	switch (pinfo->client_addr.ss.ss_family) {
		case AF_INET:
			bind(fd, (const struct sockaddr *) &pinfo->server_addr, sizeof(struct sockaddr_in));
			connect(fd, (struct sockaddr *) &pinfo->client_addr, sizeof(struct sockaddr_in));
			break;
		case AF_INET6:
			LOG("BIG NO");
			setsockopt(fd, IPPROTO_IPV6, IPV6_V6ONLY, (const char *)&off, sizeof(off));
			bind(fd, (const struct sockaddr *) &pinfo->server_addr, sizeof(struct sockaddr_in6));
			connect(fd, (struct sockaddr *) &pinfo->client_addr, sizeof(struct sockaddr_in6));
			break;
		default:
			OPENSSL_assert(0);
			break;
	}

	/* Set new fd and set BIO to connected */
	BIO_set_fd(SSL_get_rbio(ssl), fd, BIO_NOCLOSE);
	//BIO_ctrl(SSL_get_rbio(ssl), BIO_CTRL_DGRAM_SET_CONNECTED, 0, &pinfo->client_addr.ss);

	/* Finish handshake */
	do { ret = SSL_accept(ssl); }
	while (ret == 0);
	if (ret < 0) {
		perror("SSL_accept");
		LOG("%s\n", ERR_error_string(ERR_get_error(), buf));
		goto cleanup;
	}

	/* Set and activate timeouts */
	timeout.tv_sec = 5;
	timeout.tv_usec = 0;
	BIO_ctrl(SSL_get_rbio(ssl), BIO_CTRL_DGRAM_SET_RECV_TIMEOUT, 0, &timeout);

	if (verbose) {
		if (pinfo->client_addr.ss.ss_family == AF_INET) {
			LOG("\nThread %lx: accepted connection from %s:%d\n",
			        id_function(),
			        inet_ntop(AF_INET, &pinfo->client_addr.s4.sin_addr, addrbuf, INET6_ADDRSTRLEN),
			        ntohs(pinfo->client_addr.s4.sin_port));
			LOG("\nThread %lx: using socket %s:%d\n",
			        id_function(),
			        inet_ntop(AF_INET, &pinfo->server_addr.s4.sin_addr, addrbuf, INET6_ADDRSTRLEN),
			        ntohs(pinfo->server_addr.s4.sin_port));
		} else {
			LOG("\nThread %lx: accepted connection from %s:%d\n",
			        id_function(),
			        inet_ntop(AF_INET6, &pinfo->client_addr.s6.sin6_addr, addrbuf, INET6_ADDRSTRLEN),
			        ntohs(pinfo->client_addr.s6.sin6_port));
			LOG("\nThread %lx: using socket %s:%d\n",
			        id_function(),
			        inet_ntop(AF_INET, &pinfo->server_addr.s6.sin6_addr, addrbuf, INET6_ADDRSTRLEN),
			        ntohs(pinfo->server_addr.s6.sin6_port));
		}
	}

	// aqua
	struct node_info *	node;
	int			node_i	= get_free_node(state); // FIXME: remove null.

	if (node_i == -1)
		goto cleanup;

	node		= &state->nodes[node_i];
	node->used	= 1;
	node->fd	= fd;
	node->peer_addr	= pinfo->client_addr;
	node->our_addr	= pinfo->server_addr;
	node->read_tid	= pinfo->tid;
	node->ssl	= ssl;
	LOG("init mutex node %i", node->index);
	//pthread_mutex_init(&node->queue_lock, NULL);
	//pthread_mutex_init(&node->ssl_lock, NULL);

	if (veryverbose && SSL_get_peer_certificate(ssl)) {
		LOG("------------------------------------------------------------\n");
		X509_NAME_print_ex_fp(stdout, X509_get_subject_name(SSL_get_peer_certificate(ssl)),
							  1, XN_FLAG_MULTILINE);
		LOG("\n\n Cipher: %s", SSL_CIPHER_get_name(SSL_get_current_cipher(ssl)));
		LOG("\n------------------------------------------------------------\n\n");
	}

	create_rated_write_thread(node);

	fd_set		read_fds;
	int		ready;

	FD_ZERO(&read_fds);
	FD_SET(fd, &read_fds);


	//LOG("sending new client notif");
	//send_new_client(node); //FIXME if !0 goto err

	while ((ready = select(fd + 1, &read_fds, NULL, NULL, NULL)))
	{
		if (ready == -1)
		{
			perror("select()");
			goto cleanup;
		}
		MUTEX_LOCK(&node->ssl_lock);
		if ((SSL_get_shutdown(ssl) & SSL_RECEIVED_SHUTDOWN) || num_timeouts >= max_timeouts)
		{
			MUTEX_UNLOCK(&node->ssl_lock);
			break;
		}
		//MUTEX_UNLOCK(&node->ssl_lock);

		reading = 1;
		while (reading) {
			//MUTEX_LOCK(&node->ssl_lock);
			len = SSL_read(ssl, buf, sizeof(buf));
			process_fastpath((uint8_t *) buf, len, node);

			switch (SSL_get_error(ssl, len)) {
				case SSL_ERROR_NONE:
					LOG("no error\n");
					// if (verbose) {
					// 	LOG("Thread %lx: read %d bytes\n", id_function(), (int) len);
					// 	if (pinfo->client_addr.ss.ss_family == AF_INET) {
					// 		LOG("\nThread %lx: read from connection from %s:%d\n",
					// 		    id_function(),
					// 		    inet_ntop(AF_INET, &pinfo->client_addr.s4.sin_addr, addrbuf, INET6_ADDRSTRLEN),
					// 		    ntohs(pinfo->client_addr.s4.sin_port));
					// 	} else {
					// 		LOG("\nThread %lx: read from connection from %s:%d\n",
					// 		    id_function(),
					// 		    inet_ntop(AF_INET6, &pinfo->client_addr.s6.sin6_addr, addrbuf, INET6_ADDRSTRLEN),
					// 		    ntohs(pinfo->client_addr.s6.sin6_port));
					// 	}
					// }
					reading = 0;
					break;
				case SSL_ERROR_WANT_READ:
					/* Handle socket timeouts */
					//!
					LOG("want read\n");

					if (BIO_ctrl(SSL_get_rbio(ssl), BIO_CTRL_DGRAM_GET_RECV_TIMER_EXP, 0, NULL)) {
						num_timeouts++;
						reading = 0;
					}
					/* Just try again */
					break;
				case SSL_ERROR_ZERO_RETURN:
					reading = 0;
					break;
				case SSL_ERROR_SYSCALL:
					LOG("Socket read error: ");
					if (!handle_socket_error())
						goto cleanup;
					reading = 0;
					break;
				case SSL_ERROR_SSL:
					LOG("SSL read error: ");
					LOG("%s (%d)\n", ERR_error_string(ERR_get_error(), buf), SSL_get_error(ssl, len));
					MUTEX_UNLOCK(&node->ssl_lock);
					goto cleanup;
					break;
				default:
					LOG("Unexpected error while reading!\n");
					MUTEX_UNLOCK(&node->ssl_lock);
					goto cleanup;
					break;
			}
			MUTEX_UNLOCK(&node->ssl_lock);
		}
	}
	LOG("lost connection");

	pthread_cancel(node->write_tid);
	SSL_shutdown(ssl);

cleanup:
#ifdef WIN32
	closesocket(fd);
#else
	close(fd);
#endif
	free(info);
	SSL_free(ssl);
	ERR_remove_state(0);
	if (verbose)
		LOG("Thread %lx: done, connection closed.\n", id_function());
#if WIN32
	ExitThread(0);
#else
	pthread_exit( (void *) NULL );
#endif
}

#ifdef WIN32
DWORD WINAPI
start_server(LPVOID *info)
#else
void *
start_server(void *	state_info _unused)
#endif
{
	int		port	= state->aqua_dtls_port;
	const char *	local_address = "";
	int		fd;
	addr server_addr, client_addr;
	SSL_CTX *ctx;
	SSL *ssl;
	BIO *bio;
	struct timeval timeout;
	struct pass_info *info;
	const int on = 1, off = 0;

	memset(&server_addr, 0, sizeof(struct sockaddr_storage));

	// FIXME: tmp for no AF_inet6
//	if (strlen(local_address) == 0) {
//		server_addr.s6.sin6_family = AF_INET6;
//#ifdef HAVE_SIN6_LEN
//		server_addr.s6.sin6_len = sizeof(struct sockaddr_in6);
//#endif
//		server_addr.s6.sin6_addr = in6addr_any;
//		server_addr.s6.sin6_port = htons(port);
//	} else {
//		if (inet_pton(AF_INET, local_address, &server_addr.s4.sin_addr) == 1) {
			server_addr.s4.sin_family = AF_INET;
#ifdef HAVE_SIN_LEN
			server_addr.s4.sin_len = sizeof(struct sockaddr_in);
#endif
			server_addr.s4.sin_port = htons(port);
//		} else if (inet_pton(AF_INET6, local_address, &server_addr.s6.sin6_addr) == 1) {
//			server_addr.s6.sin6_family = AF_INET6;
//#ifdef HAVE_SIN6_LEN
//			server_addr.s6.sin6_len = sizeof(struct sockaddr_in6);
//#endif
//			server_addr.s6.sin6_port = htons(port);
//		} else {
//			goto error;
//		}
//	}

	THREAD_setup();
	OpenSSL_add_ssl_algorithms();
	SSL_load_error_strings();
	ctx = SSL_CTX_new(DTLSv1_server_method());
	/* We accept all ciphers, including NULL.
	 * Not recommended beyond testing and debugging
	 */
	SSL_CTX_set_cipher_list(ctx, SERVER_CIPHER_LIST);
	//SSL_CTX_set_cipher_list(ctx, "ALL:NULL:eNULL:aNULL");
	SSL_CTX_set_session_cache_mode(ctx, SSL_SESS_CACHE_OFF);

	if (!SSL_CTX_use_certificate_file(ctx, (char *) state->cert_file, SSL_FILETYPE_PEM))
		LOG("\nERROR: no certificate found!");

	if (!SSL_CTX_use_PrivateKey_file(ctx, (char *) state->key_file, SSL_FILETYPE_PEM))
		LOG("\nERROR: no private key found!");

	if (!SSL_CTX_check_private_key (ctx))
		LOG("\nERROR: invalid private key!");

	/* Client has to authenticate */
	SSL_CTX_set_verify(ctx, SSL_VERIFY_NONE, dtls_verify_callback);

	SSL_CTX_set_read_ahead(ctx, 1);
	SSL_CTX_set_cookie_generate_cb(ctx, generate_cookie);
	SSL_CTX_set_cookie_verify_cb(ctx, verify_cookie);

#if WIN32
	WSAStartup(MAKEWORD(2, 2), &wsaData);
#endif
	fd = socket(server_addr.ss.ss_family, SOCK_DGRAM, 0);
	if (fd < 0) {
		perror("socket");
		exit(-1);
	}

#ifdef WIN32
	setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (const char*) &on, (socklen_t) sizeof(on));
#else
	setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (const void*) &on, (socklen_t) sizeof(on));
#ifdef SO_REUSEPORT
	setsockopt(fd, SOL_SOCKET, SO_REUSEPORT, (const void*) &on, (socklen_t) sizeof(on));
#endif
#endif

	if (server_addr.ss.ss_family == AF_INET) {
		bind(fd, (const struct sockaddr *) &server_addr, sizeof(struct sockaddr_in));
	} else {
		LOG("BIG NO");
		setsockopt(fd, IPPROTO_IPV6, IPV6_V6ONLY, (const char *) &off, sizeof(off));
		bind(fd, (const struct sockaddr *) &server_addr, sizeof(struct sockaddr_in6));
	}

	while (1) {
		memset(&client_addr, 0, sizeof(struct sockaddr_storage));

		/* Create BIO */
		bio = BIO_new_dgram(fd, BIO_NOCLOSE);

		/* Set and activate timeouts */
		timeout.tv_sec = 5;
		timeout.tv_usec = 0;
		BIO_ctrl(bio, BIO_CTRL_DGRAM_SET_RECV_TIMEOUT, 0, &timeout);

		ssl = SSL_new(ctx);

		SSL_set_bio(ssl, bio, bio);
		SSL_set_options(ssl, SSL_OP_COOKIE_EXCHANGE);

		while (DTLSv1_listen(ssl, &client_addr) <= 0)
			;

		info = (struct pass_info*) malloc (sizeof(struct pass_info));
		memcpy(&info->server_addr, &server_addr, sizeof(struct sockaddr_storage));
		memcpy(&info->client_addr, &client_addr, sizeof(struct sockaddr_storage));
		info->ssl = ssl;

#ifdef WIN32
		if (CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE) connection_handle, info, 0, &indo->tid) == NULL) {
			exit(-1);
		}
#else
		if (pthread_create(&info->tid, NULL, connection_handle, info) != 0) {
			perror("pthread_create");
			exit(-1);
		}
#endif
	}
error:

	THREAD_cleanup();
#ifdef WIN32
	ExitThread(0);
#else
	pthread_exit((void *) NULL);
#endif
}

#ifdef WIN32
DWORD WINAPI
start_client(LPVOID *info)
#else
void *
start_client(void *	node_info)
#endif
{
	struct node_info *	node = node_info;
	int fd;
	addr remote_addr, local_addr;
	uint8_t buf[BUFFER_SIZE];
	char addrbuf[INET6_ADDRSTRLEN];
	socklen_t len;
	SSL_CTX *ctx;
	SSL *ssl;
	BIO *bio;
	int reading = 0;
	struct timeval timeout;
 	int num_timeouts = 0, max_timeouts = 5; // FIXME put this in config somewhere.
#if WIN32
	WSADATA wsaData;
#endif

////	memset((void *) &remote_addr, 0, sizeof(struct sockaddr_storage));
	memcpy((void *) &remote_addr, &node->peer_addr, sizeof(struct sockaddr_storage));
	memset((void *) &local_addr, 0, sizeof(struct sockaddr_storage));
	local_addr.ss.ss_family	= node->peer_addr.ss.ss_family; // and that should do it, right? any port, any ip.
	//local_addr.ss.ss_family	= AF_INET; // and that should do it, right? any port, any ip.
	if (local_addr.ss.ss_family == AF_INET)
	{
		remote_addr.s4.sin_port = htons(remote_addr.s4.sin_port);
#ifdef HAVE_SIN_LEN
		local_addr.s4.sin_len = sizeof(struct sockaddr_in);
		remote_addr.s4.sin_len = sizeof(struct sockaddr_in);
#endif
	}
	else
	{

		LOG("BIG NO");
		remote_addr.s6.sin6_port = htons(remote_addr.s6.sin6_port);
#ifdef HAVE_SIN6_LEN
		local_addr.s6.sin6_len = sizeof(struct sockaddr_in6);
		remote_addr.s6.sin6_len = sizeof(struct sockaddr_in6);
#endif
	}

////	if (inet_pton(AF_INET, remote_address, &remote_addr.s4.sin_addr) == 1) {
////		remote_addr.s4.sin_family = AF_INET;
////#ifdef HAVE_SIN_LEN
////		remote_addr.s4.sin_len = sizeof(struct sockaddr_in);
////#endif
////		remote_addr.s4.sin_port = htons(port);
////	} else if (inet_pton(AF_INET6, remote_address, &remote_addr.s6.sin6_addr) == 1) {
////		remote_addr.s6.sin6_family = AF_INET6;
////#ifdef HAVE_SIN6_LEN
////		remote_addr.s6.sin6_len = sizeof(struct sockaddr_in6);
////#endif
////		remote_addr.s6.sin6_port = htons(port);
////	} else {
////		return;
////	}

#ifdef WIN32
	WSAStartup(MAKEWORD(2, 2), &wsaData);
#endif

	fd = socket(node->peer_addr.ss.ss_family, SOCK_DGRAM, 0);
	if (fd < 0) {
		perror("socket");
		exit(-1);
	}

	if (local_addr.ss.ss_family == AF_INET)
		bind(fd, (const struct sockaddr *) &local_addr, sizeof (struct sockaddr_in));
	else
	{
		bind(fd, (const struct sockaddr *) &local_addr, sizeof (struct sockaddr_in6));
			LOG("BIG NO");
	}

	ctx = SSL_CTX_new(DTLSv1_client_method());
	SSL_CTX_set_cipher_list(ctx, SERVER_CIPHER_LIST);
	//SSL_CTX_set_cipher_list(ctx, "ALL:NULL:eNULL:aNULL");
	//SSL_CTX_set_cipher_list(ctx, "eNULL:!MD5");

	if (!SSL_CTX_use_certificate_file(ctx, (char *) state->cert_file, SSL_FILETYPE_PEM))
		LOG("\nERROR: no certificate found!");

	if (!SSL_CTX_use_PrivateKey_file(ctx, (char *) state->key_file, SSL_FILETYPE_PEM))
		LOG("\nERROR: no private key found!");

	if (!SSL_CTX_check_private_key (ctx))
		LOG("\nERROR: invalid private key!");

	SSL_CTX_set_verify_depth(ctx, 0);
	SSL_CTX_set_read_ahead(ctx, 1);

	ssl = SSL_new(ctx);

	/* Create BIO, connect and set to already connected */
	bio = BIO_new_dgram(fd, BIO_CLOSE);
	if (remote_addr.ss.ss_family == AF_INET) {
		connect(fd, (struct sockaddr *) &remote_addr, sizeof(struct sockaddr_in));
	} else {
		connect(fd, (struct sockaddr *) &remote_addr, sizeof(struct sockaddr_in6));
	}
	BIO_ctrl(bio, BIO_CTRL_DGRAM_SET_CONNECTED, 0, &remote_addr.ss);

	SSL_set_bio(ssl, bio, bio);

	if (SSL_connect(ssl) < 0) {
		LOG("SSL_connect: %s", strerror(errno));
		LOG("%s\n", ERR_error_string(ERR_get_error(), (char *) buf));
		if (remote_addr.ss.ss_family == AF_INET) {
			LOG("from: 4 %s",
			         inet_ntop(AF_INET, &remote_addr.s4.sin_addr, addrbuf, INET6_ADDRSTRLEN));
		} else {

			LOG("BIG NO");
			LOG("from: 6 %s",
			         inet_ntop(AF_INET6, &remote_addr.s6.sin6_addr, addrbuf, INET6_ADDRSTRLEN));
		}
		goto cleanup_init;
	}

	/* Set and activate timeouts */
	timeout.tv_sec = 3;
	timeout.tv_usec = 0;
	BIO_ctrl(bio, BIO_CTRL_DGRAM_SET_RECV_TIMEOUT, 0, &timeout);

	if (verbose) {
		if (remote_addr.ss.ss_family == AF_INET) {
			LOG("\nConnected to %s\n",
			         inet_ntop(AF_INET, &remote_addr.s4.sin_addr, addrbuf, INET6_ADDRSTRLEN));
		} else {
			LOG("BIG NO");
			LOG("\nConnected to %s\n",
			         inet_ntop(AF_INET6, &remote_addr.s6.sin6_addr, addrbuf, INET6_ADDRSTRLEN));
		}
	}

	if (veryverbose && SSL_get_peer_certificate(ssl)) {
		LOG("------------------------------------------------------------\n");
		X509_NAME_print_ex_fp(stdout, X509_get_subject_name(SSL_get_peer_certificate(ssl)),
		                      1, XN_FLAG_MULTILINE);
		LOG("\n\n Cipher: %s", SSL_CIPHER_get_name(SSL_get_current_cipher(ssl)));
		LOG("\n------------------------------------------------------------\n\n");
	}

	// FIXME: we're going to need other things here
	node->ssl	= ssl;
	create_rated_write_thread(node);

	fd_set		read_fds;
	int		ready;

	FD_ZERO(&read_fds);
	FD_SET(fd, &read_fds);

	send_ack(node->cookie, 0, node->index);


	while ((ready = select(fd + 1, &read_fds, NULL, NULL, NULL)))
	{
		if (ready == -1)
		{
			LOG("select: %s:", strerror(errno));
			goto cleanup;
		}
		MUTEX_LOCK(&node->ssl_lock);
		if ((SSL_get_shutdown(ssl) & SSL_RECEIVED_SHUTDOWN) || num_timeouts >= max_timeouts)
		{
			MUTEX_UNLOCK(&node->ssl_lock);
			break;
		}
		//MUTEX_UNLOCK(&node->ssl_lock);
		reading = 1;
		while (reading) {
			//MUTEX_LOCK(&node->ssl_lock);
			len = SSL_read(ssl, buf, sizeof(buf));
			process_fastpath(buf, len, node);

			switch (SSL_get_error(ssl, len)) {
				case SSL_ERROR_NONE:
					//if (verbose) {
					//	//LOG("read %d bytes\n", (int) len);
					//}
					reading = 0;
					break;
				case SSL_ERROR_WANT_READ:
					/* Stop reading on socket timeout, otherwise try again */
					if (BIO_ctrl(SSL_get_rbio(ssl), BIO_CTRL_DGRAM_GET_RECV_TIMER_EXP, 0, NULL)) {
						LOG("Timeout! No response received.\n");
						reading = 0;
					}
					break;
				case SSL_ERROR_ZERO_RETURN:
					reading = 0;
					break;
				case SSL_ERROR_SYSCALL:
					LOG("Socket read error: ");
					if (!handle_socket_error()) exit(1);
					reading = 0;
					break;
				case SSL_ERROR_SSL:
					LOG("SSL read error: ");
					LOG("%s (%d)\n", ERR_error_string(ERR_get_error(), (char *) buf), SSL_get_error(ssl, len));
					exit(1);
					break;
				default:
					LOG("Unexpected error while reading!\n");
					exit(1);
					break;
			}
			MUTEX_UNLOCK(&node->ssl_lock);
		}
	}

cleanup_init:
	send_ack(node->cookie, 8, node->index);
cleanup:
	// FIXME: kill other thread, free node.

#ifdef WIN32
	closesocket(fd);
#else
	close(fd);
#endif
	if (verbose)
		LOG("Connection closed.\n");

#ifdef WIN32
	WSACleanup();
#endif

#if WIN32
	ExitThread(0);
#else
	pthread_exit((void *) NULL);
#endif
}
